"""does naive bayesian classification"""

from reverend.thomas import Bayes
import os, sys, errno, signal

HTML_TEMPLATE = sys.prefix + "/share/feedisto/feedisto.html_template"
DEBUG = False
WARNING = True

class Classifier:
	"""for rating items"""
	def __init__(self, config):
		self.config = config
		self.bayes_file = self.config.my_dir + "/bayes.bay"
		self.guesser = Bayes()
		self.train = self.guesser.train
	def save(self):
		self.guesser.save(self.bayes_file)
	def load(self):
		if os.path.exists(self.bayes_file):
			self.guesser.load(self.bayes_file)
		else:
			# the three categories are needed
			self.guesser.train('spam', '1')
			self.guesser.train('ham', '1')
			self.guesser.train('gold', '1')
	def parse(self, bunch):
		"""give a list of NewsItems to give them a relevance"""
		self.load()
		for item in bunch:
			# spam is bad
			# ham is good
			# gold is extraordinary
			list = self.guesser.guess(item.raw_text())
			if len(list) != 3:
				return bunch
			if list[0][1] == "gold":
				item.relevance = 1.0
			else:
				relevance = 0.0
				for guess in list:
					if guess[0] == "spam":
						relevance -= guess[1]
					elif guess[0] == "ham":
						relevance += guess[1]
				item.relevance = relevance
		return bunch

import socket, os, sys, shutil, datetime
class Server:
	"""starting a server to get Feedback from the User, to train the filters"""
	def __init__(self, config):
		self.config = config
		self.bayes_file = os.path.expanduser(self.config.my_dir + "/bayes.bay")
		self.template = self.config.html_template
		if not self.template:
			if os.path.exists(HTML_TEMPLATE):
				self.template = HTML_TEMPLATE
			else:
				print "No HTML Template found at ", HTML_TEMPLATE
		self.pidfile = self.config.bayes_server_pidfile
		if self.pidfile:
			self.pidfile = os.path.expanduser(self.pidfile)
		self.template = os.path.expanduser(self.template)
		self.daemonized = False
		self.pid = False
	def put(self, bunch):
		if len(bunch) == 0:
			return
		# ok now the real server stuff
		self.daemonize()
		if self.daemonized:
			self.bunch = bunch
			self.create_page(bunch)
			self.bind()
			while True:
				# there should be timeout set
				# 1 hour or so
				try:	
					(clientsock, addr) = self.socket.accept()
					self.handle(clientsock.makefile('ab'))
					clientsock.close()
				except socket.timeout:
					pass
	def train(self, fileobj):
		if DEBUG:
			print "training now!"
		# get whole header
		content_length = 0
		line = True
		while line:
			line = fileobj.readline()[:-2]
			if line[:16] == "Content-Length: ":
				content_length = int( line[16:]	)
		# now the content
		data = fileobj.read(content_length).split('&')
		#  easy, now the training
		tommy = Bayes()
		if os.path.exists(self.bayes_file):
			tommy.load(self.bayes_file)
		for thing in data:
			if len(thing) == 0 or ( not thing.find('=') ): continue
			hasch, clazz = thing.split('=')
			for item in self.bunch:
				if item.hash() == int(hasch):
					tommy.train(clazz, item.raw_text())
		tommy.save(self.bayes_file)
		if DEBUG:
			print data
			print "training ended"
		# ready, steady, go
		if self.pidfile and os.path.exists(self.pidfile):
			os.unlink(self.pidfile)
		sys.exit(1)
	def handle(self, socket):
		line = socket.readline().split(" ")
		if line[0] == "GET":
			if line[1] == "/":
				if DEBUG:
					fh = open("server.log", 'w')
					fh.write(self.page.encode("utf-8"))
					fh.close()
					print "server.log written"
				socket.write(self.page.encode("utf-8"))
				socket.flush()
			elif line[1] == "/feeds.css" and os.path.exists("data/feeds.css"):
				for cssline in open("data/feeds.css"):
					socket.write(cssline)
				socket.flush()
			else:
				if DEBUG:
					print "i don't handle request:", line
		elif line[0] == "POST":
			self.train(socket)
			socket.write(self.page.encode("utf-8"))
			socket.flush()
		else:
			if DEBUG:
				print "what's this??", line
		socket.close()
	def bind(self):
		self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		host, port = self.config.bayes_server_addr.split(":")
		port = int(port)	
		try:
			self.socket.bind((host, port))
		except socket.error:
			if WARNING:
				print "warning: binding the socket failed. Exiting server."
			sys.exit(0)
		self.socket.listen(1)
	def create_page(self, bunch):
		page = ""
		content = ""
		for line in open(self.template):
			page += line
		item_buttons  = 'Important <input type="radio" name=\"%s\" value="ham"  /><br />\n'
		item_buttons += 'Spam <input type="radio" name=\"%s\" value="spam" /><br />\n'
		one_item = "\n<tr>\n<td class=\"buttons\">"+item_buttons+"</td><td class=\"content\">\n%s\n</td>\n</tr>\n"
		content += "<form method=\"post\"><table>\n"
		for item in bunch:
			content += one_item % (item.hash(), item.hash(), item.html())
		content += "<tr><td><input type=\"submit\" value=\"train\" /></td><td></td></tr>\n"
		content += "</table>\n"
		content += "</form>\n"
		page = page.replace("#content#", content)
		page = page.replace("#title#", self.config.title)
		page = page.replace("#subtitle#", self.config.subtitle)
		date = datetime.datetime.now().strftime(" %d.%m.%Y %H:%M ") 
		page = page.replace("#date#", date)
		self.page = page
	def daemonize(self):
		# do the UNIX double-fork magic, see Stevens' "Advanced 
		# Programming in the UNIX Environment" for details (ISBN 0201563177)
		try: 
        		pid = os.fork() 
        		if pid > 0:
            			# let the rest of the program run further
				return
    		except OSError, e: 
        		print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
	        	sys.exit(1)

    		# decouple from parent environment
    		os.setsid() 

    		# do second fork
    		try: 
        		self.pid = os.fork() 
        		if self.pid > 0:
	       	     		# exit from second parent, print eventual PID before
				# this is not the daemon procees, but the parent!
        		    	print "Daemon PID %d" % self.pid 
				self.pid_file_handling()
            			sys.exit(0) 
	    	except OSError, e: 
        		print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror) 
        		sys.exit(1) 
		self.daemonized = True
		# kill an old daemon
	def pid_file_handling(self):
		"""look for a pid file to kill an old server"""
		if self.pidfile and os.path.exists(self.pidfile):
			# kill old daemon
			fh = open(self.pidfile, 'r')
			old_pid = fh.readline()[:-1]
			fh.close
			if old_pid:
				if DEBUG:
					print "found old pid:", old_pid
				old_pid = int( old_pid )
				fh.close()
				if DEBUG:
					print "killing old daemon:", old_pid
				try:
					os.kill(old_pid, 0)
				except os.error, args:
					if args[0] != errno.ESRCH: # NO SUCH PROCESS
						if DEBUG:
							print "no such process"
					else:
						if DEBUG:
							print "strange kill error:", args
				else:
					if DEBUG:
						print "kill failed, trying SIGTERM"
					os.kill(old_pid, signal.SIGTERM)
			# write the current pid to file
			fh = open(self.pidfile, 'w')
			fh.write("%i\n" % self.pid)
			fh.close()
			if DEBUG:
				print "wrote pid %s to file: %s" % ( self.pid, self.pidfile )

class main:
	def __init__(self, config):
		# act like i could do anything
		self.types = ['parser', 'output']
		self.config = config
		self.parser = Classifier(self.config)
		self.parse = self.parser.parse # just route the function
	def put(self, bunch):
		"""start the server (it will daemonize)"""
		server = Server(self.config)
		server.put(bunch)
